@make(ais)
@aiInfo(topic=<Storage allocation for constrained record subtypes>,
number=[AI-00566/08],
classcode=[ra],
statuscode=[WA],
mainref=[13.07.02@\(05)],
date=[89-10-30])
@aiHeading()
@begin(history)
@begin(toplines)
@mainref()
@standard[11.01@\(08)]
!class ramification 88-06-06
@cbon()!status approved by WG9 89-10-27@cboff()
@end(toplines)
!status ARG-approved 89-05-15 (reviewed)
!status ARG-approved (11-0-0) 88-11-01 (pending editorial review)
!status ARG-approved (12-0-0) 88-06-06 (pending editorial review)
!status work-item 88-05-09
@references[!references AI-00325, @CommentNum[83-00969]83-00969, @CommentNum[83-00972]83-00972, @CommentNum[83-01011]83-01011]
@topic{!topic Storage allocation for constrained record subtypes}
@end(history)

@subheading[!summary 89-01-05]

An implementation is expected to provide a reasonably compact representation
for values of a constrained record subtype.

@subheading[!question @cbon()89-10-30]@cboff()

Can an implementation raise STORAGE_@|ERROR for the following type declarations
or object declarations:
@begin(example)
@key[type] REC_@|1 (DISC : NATURAL) @key[is]
    @key[record]
        C1 : STRING (1..DISC);
    @key[end] @key[record];
@hinge()

OBJ_@|1 : REC_@|1 (FUNC_@|CALL);  -- STORAGE_@|ERROR if FUNC_@|CALL = 100? (no)
@hinge()

@key[type] ARR @key[is] @key[array] (POSITIVE @key[range] <>) @key[of] INTEGER;
@hinge()

@key[type] REC_@|2 @key[is]
    @key[record]
        C2 : ARR (1..FUNC_@|CALL);
    @key[end] @key[record];
@hinge()

OBJ_@|2 : REC_@|2;              -- STORAGE_@|ERROR if FUNC_@|CALL = 100? (no)
@end(example)
Note that the discriminant of REC_@|1 does not have a default expression, so
every object of type REC_@|1 must be constrained to a fixed size.  Is it
acceptable for an implementation to allocate, at compile time, the largest
possible space for OBJ_@|1 and OBJ_@|2, even though the amount of space actually
used can be significantly less, and even if an attempt to allocate the
largest potentially needed space is likely to raise STORAGE_@|ERROR at
run-time?  Note that if the largest discriminant value used for REC_@|1
subtypes is, say, 200, and similarly, if the largest upper bound for
component C2 is 200, a programmer could declare a subtype with this range and
use this subtype in the declarations of REC_@|1 and REC_@|2:
@begin(example)
@key[subtype] UPPER_@|LIMIT @key[is] INTEGER @key[range] 0 .. 200;
@hinge()

@key[type] REC_@|1A (DISC : UPPER_@|LIMIT) @key[is]
    @key[record]
        C1 : STRING (1..DISC);
    @key[end] @key[record];
@hinge()

OBJ_@|1 : REC_@|1A (FUNC_@|CALL); -- STORAGE_@|ERROR if FUNC_@|CALL = 100? (no)
@hinge()

@key[type] ARR_@|A @key[is] @key[array] (UPPER_@|LIMIT @key[range] <>) @key[of] INTEGER;
@hinge()

@key[type] REC_@|2A @key[is]
    @key[record]
        C2 : ARR_@|A (1..FUNC_@|CALL);
    @key[end] @key[record];
@hinge()

OBJ_@|2 : REC_@|2A;             -- STORAGE_@|ERROR if FUNC_@|CALL = 100?
@end(example)
STORAGE_@|ERROR will not usually be raised if an implementation allocates the
maximum size permitted by the largest discriminant or index value, i.e., 200.
Is an implementation allowed to allocate 200@multiply()8 + INTEGER'@|SIZE @cbon()bits@cboff()
for OBJ_@|1 even if FUNC_@|CALL returns a value less than 200?  Can it allocate
200@multiply()INTEGER'@|SIZE @cbon()bits@cboff() for OBJ_@|2 even if FUNC_@|CALL returns a small
value?  What is an implementation allowed to do if a bound of UPPER_@|LIMIT is
not static?

@subheading[!response @cbon()89-10-30]@cboff()

If an implementation always allocates the maximum amount of space needed for
any constrained record subtype, STORAGE_@|ERROR will often be raised when the
range of allowed discriminant values is large, particularly when a
discriminant is used within the record to constrain an array component, as in
the example in the question.  The effect of raising STORAGE_@|ERROR in such
cases is to deny the corresponding language notation to programmers.  In
principle, the issue is similar to that for case statements when the case
expression has a nonstatic integer subtype.  An implementation is not allowed
to use an implementation approach that uses an excessive amount of storage
(e.g., a jump table indexed on all integer values), since this is equivalent
to nonsupport of case statements in such cases.  Similarly, for constrained
record subtypes, an implementation must take the effort to generate reasonably
compact representations (see AI-00325).

A @dq"reasonably" compact representation is not necessarily the minimal
representation.  To improve run-time performance (in the absence of a
representation clause or pragma PACK that applies to a record type), an
implementation may increase the space allocated for a constrained record
subtype above the minimal space required.  For example, it is not unreasonable
to add space for descriptor components or for gaps that allow components to be
placed on efficiently accessed storage unit boundaries.  Consider a record
containing two components that depend on discriminants:
@begin(example)
@key[subtype] SMALL @key[is] INTEGER @key[range] 1 .. 1000;
@key[type] S (D1, D2 : SMALL) @key[is]
    @key[record]
        C3 : STRING (1 .. D1);
        C4 : STRING (1 .. D2);
    @key[end] @key[record];
@hinge()

OBJ_@|S : S (27, 33);
@end(example)
For OBJ_@|S, it would not be considered @dq"reasonable" to allocate 1000 units of
storage for component C3 just to ensure that component C4 is always at the
same offset within the record.  However, for some machine architectures, it
would be considered reasonable to position C4 so it starts on a @dq"word"
boundary.  For OBJ_@|S, this would presumably mean that one padding byte is
inserted between the end of component C3 and the beginning of component C4.
In addition, it would be considered reasonable for an implementation to
provide an implementation-dependent component that points to the beginning of
component C4.  This storage strategy would provide reasonably compact
representations for all subtypes of S, even for very small discriminant values
(e.g., S(1,1)), since it reflects a storage strategy that, in general,
makes reasonable time/space tradeoffs.

In order to provide a reasonably compact representation for each subtype, it
is not always necessary to give each subtype a different amount of storage
space.  If every subtype occupies essentially the same number of storage
units, then it could be reasonable to give each subtype the same amount of
space.  For example:
@begin(example)
@key[subtype] VERY_@|SMALL @key[is] INTEGER @key[range] 1..6;
@key[type] SV (D1, D2 : VERY_@|SMALL) @key[is]  -- assume D1 and D2 occupy 8 bytes
    @key[record]
        C3 : STRING (1 .. D1);
        C4 : STRING (1 .. D2);
    @key[end] @key[record];
@end(example)
Assume that D1 and D2 each occupy 4 bytes and that padding bits are placed
after components C3 and C4 so that the number of bytes allocated for
these components is a multiple of 4.  Further assume that no
implementation-dependent components are allocated.  Then the
minimal allocation for subtypes of SV ranges from 16 to 24 bytes.
Given the small variation in size, it would be reasonable to allocate
24 bytes for every subtype of SV, placing component C4 at a fixed storage
unit offset within the object.

In short, any storage allocation strategy is allowed as long as it provides a
reasonably compact representation for all subtypes.  With respect to the
examples given in the question, this rule means that:
@begin(itemize0)
STORAGE_@|ERROR cannot be raised if FUNC_@|CALL yields 100 and there
are approximately 100@multiply()8 + INTEGER'@|SIZE @cbon()bits@cboff() of available storage
for OBJ_@|1 (100@multiply()INTEGER'@|SIZE @cbon()bits@cboff() for OBJ_@|2).


It is not acceptable to allocate 200@multiply()INTEGER'@|SIZE @cbon()bits@cboff()
for OBJ_@|2 if FUNC_@|CALL returns a small value.


The answers are not changed if the discriminant subtype is
nonstatic.
@end(itemize0)
